@@ -3,7 +3,12 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE PatternGuards #-}
 -- | A sqlite backend for persistent.
+--
+-- Note: If you prepend @WAL=off @ to your connection string, it will disable
+-- the write-ahead log. For more information, see
+-- <https://github.com/yesodweb/persistent/issues/363>.
 module Database.Persist.Sqlite
     ( withSqlitePool
     , withSqliteConn
@@ -19,47 +24,71 @@ import Database.Persist.Sql
 import qualified Database.Sqlite as Sqlite
 
 import Control.Monad.IO.Class (MonadIO (..))
-import Control.Monad.Logger (NoLoggingT, runNoLoggingT)
+import Control.Monad.Logger (NoLoggingT, runNoLoggingT, MonadLogger)
 import Data.IORef
 import qualified Data.Map as Map
 import Control.Monad.Trans.Control (control)
+import Data.Acquire (Acquire, mkAcquire, with)
 import qualified Control.Exception as E
 import Data.Text (Text)
-import Control.Monad (mzero)
 import Data.Aeson
+import Data.Aeson.Types (modifyFailure)
 import qualified Data.Text as T
 import Data.Conduit
 import qualified Data.Conduit.List as CL
 import Control.Applicative
 import Data.Int (Int64)
-import Control.Monad ((>=>))
 import Data.Monoid ((<>))
 import Control.Monad.Trans.Control (MonadBaseControl)
-import Control.Monad.Trans.Resource (ResourceT, MonadResource, runResourceT)
+import Control.Monad.Trans.Resource (ResourceT, runResourceT)
+import Control.Monad (when)
 
-createSqlitePool :: MonadIO m => Text -> Int -> m ConnectionPool
+createSqlitePool :: (MonadIO m, MonadLogger m, MonadBaseControl IO m) => Text -> Int -> m ConnectionPool
 createSqlitePool s = createSqlPool $ open' s
 
-withSqlitePool :: (MonadBaseControl IO m, MonadIO m)
+withSqlitePool :: (MonadBaseControl IO m, MonadIO m, MonadLogger m)
                => Text
                -> Int -- ^ number of connections to open
                -> (ConnectionPool -> m a) -> m a
 withSqlitePool s = withSqlPool $ open' s
 
-withSqliteConn :: (MonadBaseControl IO m, MonadIO m)
-               => Text -> (Connection -> m a) -> m a
+withSqliteConn :: (MonadBaseControl IO m, MonadIO m, MonadLogger m)
+               => Text -> (SqlBackend -> m a) -> m a
 withSqliteConn = withSqlConn . open'
 
-open' :: Text -> IO Connection
-open' = Sqlite.open >=> wrapConnection
+open' :: Text -> LogFunc -> IO SqlBackend
+open' connStr logFunc = do
+    let (connStr', enableWal) = case () of
+          ()
+            | Just cs <- T.stripPrefix "WAL=on "  connStr -> (cs, True)
+            | Just cs <- T.stripPrefix "WAL=off " connStr -> (cs, False)
+            | otherwise                                   -> (connStr, True)
+
+    conn <- Sqlite.open connStr'
+    wrapConnectionWal enableWal conn logFunc
 
 -- | Wrap up a raw 'Sqlite.Connection' as a Persistent SQL 'Connection'.
 --
 -- Since 1.1.5
-wrapConnection :: Sqlite.Connection -> IO Connection
-wrapConnection conn = do
+wrapConnection :: Sqlite.Connection -> LogFunc -> IO SqlBackend
+wrapConnection = wrapConnectionWal True
+
+-- | Allow control of WAL settings when wrapping
+wrapConnectionWal :: Bool -- ^ enable WAL?
+                  -> Sqlite.Connection
+                  -> LogFunc
+                  -> IO SqlBackend
+wrapConnectionWal enableWal conn logFunc = do
+    when enableWal $ do
+        -- Turn on the write-ahead log
+        -- https://github.com/yesodweb/persistent/issues/363
+        turnOnWal <- Sqlite.prepare conn "PRAGMA journal_mode=WAL;"
+        _ <- Sqlite.step turnOnWal
+        Sqlite.reset conn turnOnWal
+        Sqlite.finalize turnOnWal
+
     smap <- newIORef $ Map.empty
-    return Connection
+    return SqlBackend
         { connPrepare = prepare' conn
         , connStmtMap = smap
         , connInsertSql = insertSql'
@@ -72,6 +101,7 @@ wrapConnection conn = do
         , connNoLimit = "LIMIT -1"
         , connRDBMS = "sqlite"
         , connLimitOffset = decorateSQLWithLimitOffset "LIMIT -1"
+        , connLogFunc = logFunc
         }
   where
     helper t getter = do
@@ -104,7 +134,7 @@ prepare' conn sql = do
         , stmtQuery = withStmt' conn stmt
         }
 
-insertSql' :: EntityDef SqlType -> [PersistValue] -> InsertSqlResult
+insertSql' :: EntityDef -> [PersistValue] -> InsertSqlResult
 insertSql' ent vals =
   case entityPrimary ent of
     Just _ ->
@@ -121,7 +151,13 @@ insertSql' ent vals =
     Nothing ->
       ISRInsertGet ins sel
         where
-          sel = "SELECT last_insert_rowid()"
+          sel = T.concat
+              [ "SELECT "
+              , escape $ fieldDB (entityId ent)
+              , " FROM "
+              , escape $ entityDB ent
+              , " WHERE _ROWID_=last_insert_rowid()"
+              ]
           ins = T.concat
               [ "INSERT INTO "
               , escape $ entityDB ent
@@ -143,15 +179,16 @@ execute' conn stmt vals = flip finally (liftIO $ Sqlite.reset conn stmt) $ do
     Sqlite.changes conn
 
 withStmt'
-          :: MonadResource m
+          :: MonadIO m
           => Sqlite.Connection
           -> Sqlite.Statement
           -> [PersistValue]
-          -> Source m [PersistValue]
-withStmt' conn stmt vals = bracketP
-    (Sqlite.bind stmt vals >> return stmt)
-    (Sqlite.reset conn)
-    (const pull)
+          -> Acquire (Source m [PersistValue])
+withStmt' conn stmt vals = do
+    _ <- mkAcquire
+        (Sqlite.bind stmt vals >> return stmt)
+        (Sqlite.reset conn)
+    return pull
   where
     pull = do
         x <- liftIO $ Sqlite.step stmt
@@ -170,22 +207,20 @@ showSqlType SqlReal = "REAL"
 showSqlType (SqlNumeric precision scale) = T.concat [ "NUMERIC(", T.pack (show precision), ",", T.pack (show scale), ")" ]
 showSqlType SqlDay = "DATE"
 showSqlType SqlTime = "TIME"
-showSqlType SqlDayTimeZoned = "TIMESTAMP"
 showSqlType SqlDayTime = "TIMESTAMP"
 showSqlType SqlBlob = "BLOB"
 showSqlType SqlBool = "BOOLEAN"
 showSqlType (SqlOther t) = t
 
-migrate' :: [EntityDef a]
+migrate' :: [EntityDef]
          -> (Text -> IO Statement)
-         -> EntityDef SqlType
+         -> EntityDef
          -> IO (Either [Text] [(Bool, Text)])
 migrate' allDefs getter val = do
     let (cols, uniqs, _) = mkColumns allDefs val
     let newSql = mkCreateTable False def (filter (not . safeToRemove val . cName) cols, uniqs)
     stmt <- getter "SELECT sql FROM sqlite_master WHERE type='table' AND name=?"
-    oldSql' <- runResourceT
-             $ stmtQuery stmt [PersistText $ unDBName table] $$ go
+    oldSql' <- with (stmtQuery stmt [PersistText $ unDBName table]) ($$ go)
     case oldSql' of
         Nothing -> return $ Right [(False, newSql)]
         Just oldSql -> do
@@ -206,23 +241,23 @@ migrate' allDefs getter val = do
 
 -- | Check if a column name is listed as the "safe to remove" in the entity
 -- list.
-safeToRemove :: EntityDef a -> DBName -> Bool
+safeToRemove :: EntityDef -> DBName -> Bool
 safeToRemove def (DBName colName)
     = any (elem "SafeToRemove" . fieldAttrs)
-    $ filter ((== (DBName colName)) . fieldDB)
+    $ filter ((== DBName colName) . fieldDB)
     $ entityFields def
 
-getCopyTable :: [EntityDef a]
+getCopyTable :: [EntityDef]
              -> (Text -> IO Statement)
-             -> EntityDef SqlType
+             -> EntityDef
              -> IO [(Bool, Text)]
-getCopyTable allDefs getter val = do
+getCopyTable allDefs getter def = do
     stmt <- getter $ T.concat [ "PRAGMA table_info(", escape table, ")" ]
-    oldCols' <- runResourceT $ stmtQuery stmt [] $$ getCols
+    oldCols' <- with (stmtQuery stmt []) ($$ getCols)
     let oldCols = map DBName $ filter (/= "id") oldCols' -- need to update for table id attribute ?
     let newCols = filter (not . safeToRemove def) $ map cName cols
     let common = filter (`elem` oldCols) newCols
-    let id_ = entityID val
+    let id_ = fieldDB (entityId def)
     return [ (False, tmpSql)
            , (False, copyToTemp $ id_ : common)
            , (common /= filter (not . safeToRemove def) oldCols, dropOld)
@@ -231,8 +266,6 @@ getCopyTable allDefs getter val = do
            , (False, dropTmp)
            ]
   where
-
-    def = val
     getCols = do
         x <- CL.head
         case x of
@@ -243,7 +276,7 @@ getCopyTable allDefs getter val = do
             Just y -> error $ "Invalid result from PRAGMA table_info: " ++ show y
     table = entityDB def
     tableTmp = DBName $ unDBName table <> "_backup"
-    (cols, uniqs, _) = mkColumns allDefs val
+    (cols, uniqs, _) = mkColumns allDefs def
     cols' = filter (not . safeToRemove def . cName) cols
     newSql = mkCreateTable False def (cols', uniqs)
     tmpSql = mkCreateTable True def { entityDB = tableTmp } (cols', uniqs)
@@ -268,10 +301,10 @@ getCopyTable allDefs getter val = do
         , escape tableTmp
         ]
 
-mkCreateTable :: Bool -> EntityDef a -> ([Column], [UniqueDef]) -> Text
+mkCreateTable :: Bool -> EntityDef -> ([Column], [UniqueDef]) -> Text
 mkCreateTable isTemp entity (cols, uniqs) =
   case entityPrimary entity of
-    Just _ ->
+    Just pdef ->
        T.concat
         [ "CREATE"
         , if isTemp then " TEMP" else ""
@@ -281,7 +314,7 @@ mkCreateTable isTemp entity (cols, uniqs) =
         , T.drop 1 $ T.concat $ map sqlColumn cols
         , ", PRIMARY KEY "
         , "("
-        , T.intercalate "," $ map (escape . fieldDB) $ entityFields entity
+        , T.intercalate "," $ map (escape . fieldDB) $ compositeFields pdef
         , ")"
         , ")"
         ]
@@ -291,13 +324,21 @@ mkCreateTable isTemp entity (cols, uniqs) =
         , " TABLE "
         , escape $ entityDB entity
         , "("
-        , escape $ entityID entity
-        , " INTEGER PRIMARY KEY"
+        , escape $ fieldDB (entityId entity)
+        , " "
+        , showSqlType $ fieldSqlType $ entityId entity
+        ," PRIMARY KEY"
+        , mayDefault $ defaultAttribute $ fieldAttrs $ entityId entity
         , T.concat $ map sqlColumn cols
         , T.concat $ map sqlUnique uniqs
         , ")"
         ]
 
+mayDefault :: Maybe Text -> Text
+mayDefault def = case def of
+    Nothing -> ""
+    Just d -> " DEFAULT " <> d
+
 sqlColumn :: Column -> Text
 sqlColumn (Column name isNull typ def _cn _maxLen ref) = T.concat
     [ ","
@@ -305,9 +346,7 @@ sqlColumn (Column name isNull typ def _cn _maxLen ref) = T.concat
     , " "
     , showSqlType typ
     , if isNull then " NULL" else " NOT NULL"
-    , case def of
-        Nothing -> ""
-        Just d -> " DEFAULT " <> d
+    , mayDefault def
     , case ref of
         Nothing -> ""
         Just (table, _) -> " REFERENCES " <> escape table
@@ -334,17 +373,19 @@ escape (DBName s) =
 data SqliteConf = SqliteConf
     { sqlDatabase :: Text
     , sqlPoolSize :: Int
-    }
+    } deriving Show
 
+instance FromJSON SqliteConf where
+    parseJSON v = modifyFailure ("Persistent: error loadomg Sqlite conf: " ++) $
+      flip (withObject "SqliteConf") v $ \o -> SqliteConf
+        <$> o .: "database"
+        <*> o .: "poolsize"
 instance PersistConfig SqliteConf where
     type PersistConfigBackend SqliteConf = SqlPersistT
     type PersistConfigPool SqliteConf = ConnectionPool
-    createPoolConfig (SqliteConf cs size) = createSqlitePool cs size
+    createPoolConfig (SqliteConf cs size) = runNoLoggingT $ createSqlitePool cs size -- FIXME
     runPool _ = runSqlPool
-    loadConfig (Object o) =
-        SqliteConf <$> o .: "database"
-                   <*> o .: "poolsize"
-    loadConfig _ = mzero
+    loadConfig = parseJSON
 
 finally :: MonadBaseControl IO m
         => m a -- ^ computation to run first
